"""
Awesome new module - WILL be ignored in favor of __init__.py
"""


def new(name: str) -> None:
    """
    Creates a new service - this is from new.py

    Args:

        name (str): Name of project (feature command)

    Return: None
    """
    print(f"This is a callable as a command {name}")
