def make(env: str, is_new: bool) -> None:
    """
    make the env if it is a new one

    Args:
        env (str): name of the env
        is_new (str): if it is a new env or not

    Return: None
    """
    print(f"Initializing the {env} if it is a {is_new}")
