"""
Should be ignored
"""


def shutdown(name: str, environment: str) -> None:
    """
    Shutdown a service - should be ignored

    Args:

        name (str): Name of project
        environment (str): Name of the env

    Return: None
    """
    print(f"Shutdown a module as feature {name} {environment}")
