def unknown_world(name: str, address: str) -> None:
    """
    This is something unknown
    Args:
          name (str): Name of the unknown thing
          address (str): Address of the unknown thing

    Return: None
    """
    print(f"Unknowns {name} {address}")
